/*******************************************************************************

     Copyright (c) 2014-2018, Analog Devices, Inc.  All rights reserved.

     Redistribution and use in source and binary forms, with or without
     modification, are permitted (subject to the limitations in the
     disclaimer below) provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the
       distribution.

     * Neither the name of Analog Devices, Inc.  nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

     NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE
     GRANTED BY THIS LICENSE.  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT
     HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
     WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
     LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
     CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
     SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
     BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
     WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
     OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
     IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



*******************************************************************************/


/*******************************************************************************

    File name:    sc5xx_init.c
    Purpose:      Initialization code for ADSP-SC5xx parts.
    Notes:        Uses the power service to change processor mode and
                  frequency settings.

                  ADSP-SC5xx and ADSP-215xx parts processor speeds can be
                  varied as detailed below.

                  - By default, the initializations are for running at 400MHz
                  if DDR2 is used or 450MHz if DDR3 is used. The default
                  processor clock speed and DDR clock speed can be overridden
                  by defining the macros CONFIG_CORE_CLOCK_SPEED and
                  CONFIG_DDR_CLOCK_SPEED respectively in config.h. See
                  sc5xx_init.h and config.h for more information.

                  If the selected processor and DDR speeds are different, CGU0
                  will be used to configure the processor clock and CGU1 will
                  be used to configure the DDR clock.

*******************************************************************************/

#include "sc5xx_init.h"
#include "adi_initialize.h"

#include <services/pwr/adi_pwr.h>
#include <sys/platform.h>
#if defined(__ADSPSHARC__)
  #include <builtins.h>
  #include <sharc_rom.h>
#else
  #include <arm_rom.h>
#endif
#include <stdint.h>
#include <stdbool.h>

static bool bError = false;

static volatile uint32_t callbackEvents; /* modified in a callback therefore volatile */

static void CheckResult(ADI_PWR_RESULT result, const int32_t lineErrPos);
static void setDefaultPower(uint32_t cgu, uint32_t msel);
void initcode(ADI_ROM_BOOT_CONFIG* pBootStruct /*unused*/);

/*
 * Note errors if result code is not success.
 */
static inline void CheckResult(ADI_PWR_RESULT result, const int32_t lineErrPos)
{
    static int32_t errCount = 0;
    static int32_t lastErrLinePos __attribute__((unused));
    if (result != ADI_PWR_SUCCESS) {
        bError = true;
        errCount++;
        lastErrLinePos = lineErrPos;
    }
}

/*
 * Reset processor to default power settings
 */
static void setDefaultPower(uint32_t cgu, uint32_t msel)
{
    ADI_PWR_RESULT result;

    result = adi_pwr_SetClkControlRegister(cgu, ADI_PWR_CLK_CTL_MSEL, msel);
    CheckResult(result, __LINE__);

    result = adi_pwr_SetClkControlRegister(cgu, ADI_PWR_CLK_CTL_DF, DF_DEFAULT);
    CheckResult(result, __LINE__);
    result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_SYSSEL, SYSSEL_DEFAULT);
    CheckResult(result, __LINE__);
    result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_S0SEL, S0SEL_DEFAULT);
    CheckResult(result, __LINE__);
    result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_S1SEL, S1SEL_DEFAULT);
    CheckResult(result, __LINE__);
    result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_CSEL, CSEL_DEFAULT);
    CheckResult(result, __LINE__);
    result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_DSEL, DSEL_DEFAULT);
    CheckResult(result, __LINE__);
} /* setDefaultPower */

/*
 * Power event Callback function
 */
static void PWRCallback(void *pCBParam, uint32_t Event, void *pArg)
{
    uint32_t sclk __attribute__((unused));
    ADI_PWR_MODE  mode __attribute__((unused));

    switch ( (ADI_PWR_EVENT)Event )
    {
    case ADI_PWR_EVENT_FREQ_PRE_CHANGE:
        sclk = *(uint32_t*)(pArg);
        callbackEvents++;
        break;
    case ADI_PWR_EVENT_FREQ_POST_CHANGE:
        sclk = *(uint32_t*)(pArg);
        callbackEvents++;
        break;
    case ADI_PWR_EVENT_MODE_PRE_CHANGE:
        mode = (ADI_PWR_MODE)pArg;
        callbackEvents++;
        break;
    case ADI_PWR_EVENT_MODE_POST_CHANGE:
        mode = (ADI_PWR_MODE)pArg;
        callbackEvents++;
        break;
    default:
        break;
    }
} /* PWRCallback */

#if CONFIG_SMPU
static void smpu_config(void)
{
    /* Use SMPU instances to disable accesses to system memory that may not be
     * populated or needs to be initialized before being accessed. This will
     * avoid the possibility of an infinite stall in the system due to a
     * speculative access to a disabled or uninitialized memory. This is also
     * part of the workaround for silicon anomaly 20000018.
     */

    /* SMC - SMPU instance 0 */
    *pREG_SMPU0_CTL |= ENUM_SMPU_CTL_RSDIS;    /* Read Speculation Disable */
    *pREG_SMPU0_SECURECTL = /*0xf01*/
                     (ENUM_SMPU_SECURECTL_WSECDIS  /* Disable secure write transactions */
                    | ENUM_SMPU_SECURECTL_WNSEN    /* Enable non-secure writes */
                    | ENUM_SMPU_SECURECTL_RSECDIS  /* Disable secure read transactions */
                    | ENUM_SMPU_SECURECTL_RNSEN    /* Enable non-secure read transactions */
                    | ENUM_SMPU_SECURECTL_SBEDIS); /* Disable bus error */

} /* smpu_config */
#endif

void initcode(ADI_ROM_BOOT_CONFIG* pBootStruct /*unused*/)
{
    uint32_t fcclk = 0uL;
    uint32_t fsysclk = 0uL;
    uint32_t fsclk0 = 0uL;
    uint32_t fsclk1 = 0uL;
    uint32_t fdclk = 0uL;

    ADI_PWR_RESULT result;
    ADI_PWR_MODE mode;
    uint32_t expectedEvents = 0uL;

    bError = false;
    callbackEvents = 0uL;

    /**
     * Initialize managed drivers and/or services that have been added to
     * the project.
     * @return zero on success
     */
    adi_initComponents();

#if UNUSED_SPI_FLASH
    /* If the SPI Flash memory is not used the following is required to
     * work around silicon anomaly 20000018.
     */
    *pREG_SPI2_TXCTL = (BITM_SPI_TXCTL_TEN | BITM_SPI_TXCTL_TTI);
    *pREG_SPI2_RXCTL = (BITM_SPI_RXCTL_REN | BITM_SPI_RXCTL_RTI);
    *pREG_SPI2_CTL = (BITM_SPI_CTL_EN|BITM_SPI_CTL_MSTR| BITM_SPI_CTL_MMSE);
#endif

#if CONFIG_SMPU
    /* Configure SMPU - also required as a workaround for silicon anomaly 20000018. */
    smpu_config();
#endif

    /* Initialize the CGU to configure the required DCLK frequency. */

    result = adi_pwr_Init(CGU_DEV_0, CLKIN0);
    CheckResult(result, __LINE__);

    result = adi_pwr_InstallCallback(CGU_DEV_0, PWRCallback);
    CheckResult(result, __LINE__);

#if CONFIG_CORE_CLOCK_SPEED != CONFIG_DDR_CLOCK_SPEED
    /* If the processor and DDR are running at different speeds, use CGU1 to
    ** configure the DDR clock
    */
    result = adi_pwr_Init(CGU_DEV_1, CLKIN1);
    CheckResult(result, __LINE__);

    result = adi_pwr_InstallCallback(CGU_DEV_1, PWRCallback);
    CheckResult(result, __LINE__);

#endif

#if defined(__ADSPARM__)
    /* Set CGU clock select register to CLKO2/4 (ARM core). */
    result = adi_pwr_SetClkOutSelectRegister(CGU_DEV_0, ADI_PWR_CLKOUT_CLKO2);
#else
    /* Set CGU clock select register to CLKO0/4 (SHARC core). */
    result = adi_pwr_SetClkOutSelectRegister(CGU_DEV_0, ADI_PWR_CLKOUT_CLKO0);
#endif
    CheckResult(result, __LINE__);

    /* get the power mode */
    result = adi_pwr_GetPowerMode(CGU_DEV_0, &mode);
    CheckResult(result, __LINE__);

    if (mode != ADI_PWR_MODE_FULL_ON)
    {
        /* set power mode to full on */
        result = adi_pwr_SetPowerMode(CGU_DEV_0, ADI_PWR_MODE_FULL_ON);
        CheckResult(result, __LINE__);
        expectedEvents += 2uL; /* pre and post event */

        /* check the power mode is full on */
        result = adi_pwr_GetPowerMode(CGU_DEV_0, &mode);
        CheckResult(result, __LINE__);
        if (mode != ADI_PWR_MODE_FULL_ON)
        {
            bError = true;
        }
    }
    /* Reset processor to default power settings */
    setDefaultPower(CGU_DEV_0, MSEL_CGU_DEV_0);

#if CONFIG_CORE_CLOCK_SPEED != CONFIG_DDR_CLOCK_SPEED
    /* Get the power mode for CGU1 */
    result = adi_pwr_GetPowerMode(CGU_DEV_1, &mode);
    CheckResult(result, __LINE__);

    if (mode != ADI_PWR_MODE_FULL_ON)
    {
        /* set power mode to full on */
        result = adi_pwr_SetPowerMode(CGU_DEV_1, ADI_PWR_MODE_FULL_ON);
        CheckResult(result, __LINE__);
        expectedEvents += 2uL; /* pre and post event */

        /* check the power mode is full on */
        result = adi_pwr_GetPowerMode(CGU_DEV_1, &mode);
        CheckResult(result, __LINE__);
        if (mode != ADI_PWR_MODE_FULL_ON)
        {
            bError = true;
        }
    }
    /* Reset processor to default power settings */
    setDefaultPower(CGU_DEV_1, MSEL_CGU_DEV_1);


    /* Now configure the DDR clock using CGU1... */

    /* Set the input as CLKIN1 */
    result = adi_pwr_SelectCduClockSource(ADI_PWR_CDU_CGU_SELECT_CGU1,
                                          ADI_PWR_CDU_CLK_SELECT_CLKIN1);
    CheckResult(result, __LINE__);

    /* Set DDR clock as the output */
    result = adi_pwr_ConfigCduInputClock(ADI_PWR_CDU_CLKIN_1,
                                         ADI_PWR_CDU_CLKOUT_3);
    CheckResult(result, __LINE__);

    /* Enable the clock output */
    result = adi_pwr_EnableCduClockOutput(ADI_PWR_CDU_CLKOUT_3, true);
    CheckResult(result, __LINE__);
#endif

    result = adi_pwr_GetCoreFreq(CGU_DEV_0, &fcclk);
    CheckResult(result, __LINE__);

    result = adi_pwr_GetSystemFreq(CGU_DEV_0, &fsysclk, &fsclk0, &fsclk1);
    CheckResult(result, __LINE__);

#if CONFIG_CORE_CLOCK_SPEED != CONFIG_DDR_CLOCK_SPEED
    result = adi_pwr_GetDDRClkFreq(CGU_DEV_1, &fdclk);
#else
    result = adi_pwr_GetDDRClkFreq(CGU_DEV_0, &fdclk);
#endif
    CheckResult(result, __LINE__);

    result = adi_pwr_RemoveCallback(CGU_DEV_0, PWRCallback);
    CheckResult(result, __LINE__);
    result = adi_pwr_RemoveCallback(CGU_DEV_1, PWRCallback);
    CheckResult(result, __LINE__);

    /* check the number of callback events */
    if (expectedEvents != callbackEvents) {
        bError = true;
    }

    return;
} /* initcode */
